/*
 * Prefix Route Module
 *
 * Copyright (C) 2007 Alfred E. Heggestad
 *
 * This file is part of ser, a free SIP server.
 *
 * ser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * ser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
#include <stdlib.h>
#include "../../db/db.h"
#include "../../fifo_server.h"
#include "../../sr_module.h"
#include "../../mem/mem.h"
#include "../../data_lump_rpl.h"
#include "../../parser/parse_uri.h"
#include "../../parser/msg_parser.h"
#include "../../locking.h"
#include "../../action.h"
#include "../../route.h"
#include "tree.h"


MODULE_VERSION


/* Local variables */
static db_key_t db_cols[] = {"prefix", "route", "comment"};
static const int db_no_cols = sizeof(db_cols)/sizeof(db_cols[0]);

/* Modules parameters */
static char *db_url   = DEFAULT_DB_URL;
static char *db_table = "prefix_route";


/**
 * Load prefix rules from database and build tree in memory
 */
static int db_load(void)
{
	db_con_t *db_con = NULL;
	db_func_t dbf;
	db_res_t* db_res = NULL;
	struct tree_item *root;
	int i;
	int err;

	LOG(L_DBG, "prefix_route: db_load(): loading from database\n");

	/* Bind to mysql module  */
	if (bind_dbmod(db_url, &dbf)) {
		LOG(L_ERR, "prefix_route: db_load(): bind_dbmod failed\n");
		return -1;
	}

	if (!DB_CAPABILITY(dbf, DB_CAP_ALL)) {
		LOG(L_ERR, "prefix_route: db_load(): Database module does"
		    " not implement all functions needed by the module\n");
		return -1;
	}

	/* Open database connection */
	db_con = dbf.init(db_url);
	if (NULL == db_con) {
		LOG(L_ERR,"prefix_route: db_load(): DB connection failed\n");
		return -1;
	}

	if (dbf.use_table(db_con, db_table) < 0) {
		LOG(L_ERR, "prefix_route: db_load(): Error in use_table\n");
		err = -1;
		goto out;
	}

	/* Get all database entries */
	if ((dbf.query(db_con,      /* Database connection */
		       NULL,        /* Key names */
		       NULL,        /* Conditions */
		       NULL,        /* Values of the keys that must match */
		       db_cols,     /* Column names to return */
		       0,           /* Number of key=values pairs to compare */
		       db_no_cols,  /* Number of columns to return */
		       NULL,        /* Order by the specified column */
		       &db_res      /* Result will be stored in this variable */
		       ) != 0)
	    || (RES_ROW_N(db_res) <= 0)) {
		LOG(L_ERR, "prefix_route: db_load(): Empty database!\n");
		err = -1;
		goto out;
	}

	LOG(L_DBG, "prefix_route: db_load(): Number of rows is %d\n",
	    RES_ROW_N(db_res));

	root = tree_item_alloc();
	if (NULL == root) {
		LOG(L_ERR, "prefix_route: db_load() tree alloc failed\n");
		err = -1;
		goto out;
	}

	/* Assume Success */
	err = 0;

	/* Read from DB and build tree */
	for (i = 0; i < RES_ROW_N(db_res); i++) {
		db_val_t *val0, *val1, *val2;
		int route;

		val0 = &RES_ROWS(db_res)[i].values[0];
		val1 = &RES_ROWS(db_res)[i].values[1];
		val2 = &RES_ROWS(db_res)[i].values[2];

		if (DB_STRING != VAL_TYPE(val0) ||
		    DB_STRING != VAL_TYPE(val1) ||
		    DB_STRING != VAL_TYPE(val2)) {
			LOG(L_ERR, "prefix_route: db_load(): row %d: "
			    "expected string type\n", i);
			continue;
		}

		route = atoi(VAL_STRING(val1));

		if (route <= 0) {
			LOG(L_CRIT, "prefix_route: db_load(): invalid route %d",
			    route);
			err = -1;
			break;
		}

		if (route >= RT_NO) {
			LOG(L_CRIT, "prefix_route: db_load(): route overflow "
			    "(route=%d, max=%d)\n", route, RT_NO);
			err = -1;
			break;
		}

		if (NULL == rlist[route]) {
			LOG(L_CRIT, "prefix_route: db_load(): "
			    "route %d is not defined\n", route);
			err = -1;
			break;
		}

		if (0 != (err = tree_item_add(root, VAL_STRING(val0), route))) {
			LOG(L_CRIT, "prefix_route: db_load(): "
			    "tree_item_add() failed (%d)\n", err);
			break;
		}

		LOG(L_DBG, "  %d: prefix=%10s  route=%2s  comment=%s\n", i,
		    VAL_STRING(val0), VAL_STRING(val1), VAL_STRING(val2));
	}

	/* Error */
	if (0 != err) {
		LOG(L_ERR, "prefix_route: db_load(): error, flushing tree\n");
		tree_item_free(root);
		goto out;
	}

	/* Swap trees */
	LOG(L_DBG, "prefix_route: db_load: swapping trees\n");

	if (0 != (err = tree_swap(root)))
		goto out;


 out:
	/* Free database results */
	if (db_res && dbf.free_result(db_con, db_res) < 0) {
		LOG(L_NOTICE, "prefix_route: db_load() free_result failed\n");
	}

	/* Close database connection */
	if (NULL != db_con) {
		dbf.close(db_con);
		db_con = NULL;
	}

	return err;
}


/**
 * FIFO command - dump prefix route tree
 */
static int tree_dump(FILE *pipe, char *response_file)
{
	FILE *f = NULL;
	int err;

	f = open_reply_pipe(response_file);
	if (NULL == f) {
		LOG(L_ERR, "tree_dump: Opening of response file failed\n");
		err = -1;
		goto out;
	}

	tree_print(f);

	fprintf(f, "\n200 OK\n");

	err = 1;

 out:
	if (NULL != f)
		fclose(f);

	return err;
}


/**
 * FIFO command - reload prefix tree from database
 */
static int prefix_reload(FILE *pipe, char *response_file)
{
	LOG(L_NOTICE, "prefix_route: Reloading prefix route tree from DB\n");

	if (0 != db_load()) {
		fifo_reply(response_file,
			   "400 prefix_reload: db_load() failed\n");
		LOG(L_ERR, "prefix_route: prefix_reload(): db_load() failed\n");
	}
	else {
		fifo_reply(response_file, "200 OK\n");
	}

	return 1;
}


/**
 * Initialize module
 */
static int mod_init(void)
{
	LOG(L_NOTICE, "prefix_route: init: db_url=%s db_table=%s\n",
	    db_url, db_table);

	/* Initialise tree */
	if (0 != tree_init()) {
		LOG(L_CRIT, "prefix_route: tree_init() failed\n\n");
		return -1;
	}

	/* Register FIFO commands */
	if (register_fifo_cmd(prefix_reload, "prefix_reload", 0) < 0) {
		LOG(L_CRIT, "prefix_route: cannot register FIFO command "
		    "(prefix_reload)\n");
		return -1;
	}

	if (register_fifo_cmd(tree_dump, "prefix_dump", 0) < 0) {
		LOG(L_CRIT, "prefix_route: cannot register FIFO command "
		    "(prefix_dump)\n");
		return -1;
	}

	/* Populate database */
	if (0 != db_load()) {
		LOG(L_CRIT, "prefix_route: db_load() failed\n\n");
		return -1;
	}

	return 0;
}


/**
 * Destroy module and free all resources
 */
static void mod_destroy(void)
{
	tree_close();
}


/**
 * Extract username from the Request URI
 * First try to look at the original Request URI and if there
 * is no username use the new Request URI
 */
static int get_username(struct sip_msg* msg, str *user)
{
	if (!msg || !user)
		return -1;

	if (parse_sip_msg_uri(msg) < 0){
		LOG(L_ERR, "get_username(): bad uri\n");
		return -1; /* error, bad uri */
	}

	if (msg->parsed_uri.user.s == 0){
		/* no user in uri */
		LOG(L_ERR, "get_username(): no user in uri\n");
		return -2;
	}

	*user = msg->parsed_uri.user;

	LOG(L_INFO, "get_username: got user=(%.*s)\n",
	    user->len, user->s);

	return 0;
}


/**
 * SER Command "prefix_route"
 */
static int prefix_route(struct sip_msg *msg, char *p1, char *p2)
{
	str user;
	int err;
	int route;

	/* Unused */
	(void)p1;
	(void)p2;

	/* Get request URI */
	if (0 != (err = get_username(msg, &user))) {
		LOG(L_ERR, "prefix_route: could not get username in"
		    " Request URI (%d)\n", err);
		return err;
	}

	route = tree_route_get(&user);

	if (route > 0) {
		LOG(L_DBG, "prefix_route: found match -> route %d\n", route);
	}
	else {
		LOG(L_DBG, "prefix_route: no route found\n");
		return -1;
	}

	if (route >= RT_NO) {
		LOG(L_CRIT, "prefix_route: route overflow (route=%d, max=%d)\n",
		    route, RT_NO);
		return 1;
	}

	if (!rlist[route]) {
		LOG(L_CRIT, "prefix_route: no route %d\n", route);
		return 1;
	}

	/* If match send to route[x] */
	if ((err = run_actions(rlist[route], msg)) < 0) {
		LOG(L_ERR, "prefix_route: run_actions failed (%d)\n", err);
		return -1;
	}

	/* Success */
	return 0;
}


/*
 * Exported functions
 */
static cmd_export_t cmds[] = {
	{"prefix_route", prefix_route, 0, 0, REQUEST_ROUTE},
	{0,              0,            0, 0, 0            }
};

/*
 * Exported parameters
 */
static param_export_t params[] = {
	{"db_url",       STR_PARAM, &db_url  },
	{"db_table",     STR_PARAM, &db_table},
	{0,              0,         0        }
};

/*
 * Module description
 */
struct module_exports exports = {
	"prefix_route",  /* Module name             */
	cmds,            /* Exported functions      */
	params,          /* Exported parameters     */
	mod_init,        /* Initialization function */
	0,               /* Response function       */
	mod_destroy,     /* Destroy function        */
	0,               /* OnCancel function       */
	0                /* Child init function     */
};
